@@ -1,4 +1,4 @@
-# $Id: Client-Keepalive.pm 66 2006-05-21 20:57:06Z rcaputo $
+# $Id: Client-Keepalive.pm 73 2006-09-26 17:57:54Z rcaputo $
 
 package POE::Component::Client::Keepalive;
 
@@ -6,7 +6,7 @@ use warnings;
 use strict;
 
 use vars qw($VERSION);
-$VERSION = "0.0801";
+$VERSION = "0.0901";
 
 use Carp qw(croak);
 use Errno qw(ETIMEDOUT);
@@ -24,6 +24,23 @@ eval {
 
 use constant DEBUG => 0;
 
+# Manage connection request IDs.
+
+my $current_id = 0;
+my %active_req_ids;
+
+sub _allocate_req_id {
+  while (1) {
+    last unless exists $active_req_ids{++$current_id};
+  }
+  return $active_req_ids{$current_id} = $current_id;
+}
+
+sub _free_req_id {
+  my $id = shift;
+  delete $active_req_ids{$id};
+}
+
 # The connection manager uses a number of data structures, most of
 # them arrays.  These constants define offsets into those arrays, and
 # the comments document them.
@@ -41,6 +58,7 @@ sub SF_KEEPALIVE () {  8 }   #   $keep_alive_secs,
 sub SF_TIMEOUT   () {  9 }   #   $default_request_timeout,
 sub SF_RESOLVER  () { 10 }   #   $poco_client_dns_object,
 sub SF_SHUTDOWN  () { 11 }   #   $shutdown_flag,
+sub SF_REQ_INDEX () { 12 }   #   \%request_id_to_wheel_id,
                              # );
 
                             # $socket_xref{$socket} = [
@@ -83,6 +101,7 @@ sub RQ_START    () {  9 }   #   $request_start_time,
 sub RQ_TIMER_ID () { 10 }   #   $request_timer_id,
 sub RQ_WHEEL_ID () { 11 }   #   $request_wheel_id,
 sub RQ_ACTIVE   () { 12 }   #   $request_is_active,
+sub RQ_ID       () { 13 }   #   $request_id,
                             # ];
 
 # Create a connection manager.
@@ -143,6 +162,7 @@ sub new {
         ka_resolve_request   => "_ka_resolve_request",
         ka_add_to_queue      => "_ka_add_to_queue",
         ka_dns_response      => "_ka_dns_response",
+        ka_cancel_dns_response => "_ka_cancel_dns_response",
       },
     ],
   );
@@ -249,6 +269,7 @@ sub _ka_wake_up {
       $request,   # WHEEL_REQUEST
     ];
 
+    # store the wheel's ID in the request object
     $request->[RQ_WHEEL_ID] = $wheel->ID;
 
     # Count it as used, so we don't over commit file handles.
@@ -350,7 +371,10 @@ sub allocate {
     undef,      # RQ_TIMER_ID
     undef,      # RQ_WHEEL_ID
     1,          # RQ_ACTIVE
+    _allocate_req_id(), # RQ_ID
   ];
+  
+  $self->[SF_REQ_INDEX]{$request->[RQ_ID]} = $request;
 
   $poe_kernel->refcount_increment(
     $request->[RQ_SESSION]->ID(),
@@ -360,7 +384,74 @@ sub allocate {
   $poe_kernel->call("$self", ka_set_timeout     => $request);
   $poe_kernel->call("$self", ka_resolve_request => $request);
 
-  return;
+  return $request->[RQ_ID];
+}
+
+sub deallocate {
+  my ($self, $req_id) = @_;
+
+  croak "deallocate() requires a request ID" unless(
+    defined($req_id) and exists($active_req_ids{$req_id})
+  );
+
+  my $request = $self->[SF_REQ_INDEX]{$req_id};
+  unless (defined $request) {
+    DEBUG and warn "deallocate could not find request $req_id";
+    return;
+  }
+
+  my $conn_key = $request->[RQ_CONN_KEY];
+  my $existing_connection = $self->_check_free_pool($conn_key);
+
+  if (defined $existing_connection) {
+    # remove it from the pool, delete the socket
+    $self->_remove_socket_from_pool($existing_connection->{socket});
+    DEBUG and warn( "deallocate called, deleted already-connected socket" );
+    return;
+  }
+  else {
+    DEBUG and warn(
+      "deallocate called without an existing connection.  ",
+      "cancelling connection request"
+    );
+
+    my $heap = $poe_kernel->get_active_session->get_heap;
+    $heap->{resolve}->{$request->[RQ_ADDRESS]} = "cancelled";
+
+    $poe_kernel->call( "$self", ka_cancel_dns_response => $request );
+    return;
+  }
+}
+
+sub _ka_cancel_dns_response {
+  my ($self, $kernel, $heap, $request) = @_[OBJECT, KERNEL, HEAP, ARG0];
+
+  my $address = $request->[RQ_ADDRESS];
+  my $requests = $heap->{resolve}{$address};
+
+  # Remove the resolver request for the address of this connection
+  # request
+
+  my $req_index = @$requests;
+  while ($req_index--) {
+    next if $requests->[$req_index] eq "cancelled";
+    next unless $requests->[$req_index] == $request;
+    splice(@$requests, $req_index, 1);
+    last;
+  }
+
+  # Clean up the structure for the address if there are no more
+  # requests to resolve that address.
+
+  unless (@$requests) {
+    $heap->{resolve}{$address} = ["cancelled"];
+  }
+
+  # cancel our attempt to connect
+  $poe_kernel->alarm_remove( $request->[RQ_TIMER_ID] );
+  $poe_kernel->refcount_decrement(
+    $request->[RQ_SESSION]->ID(), "poco-client-keepalive"
+  );
 }
 
 # Set the request's timeout, in the component's context.
@@ -411,6 +502,9 @@ sub _ka_conn_failure {
   # Remove the SF_USED placeholder.
   delete $self->[SF_USED]{$wheel_id};
 
+  # remove the wheel-to-request index
+  delete $self->[SF_REQ_INDEX]{$request->[RQ_ID]};
+
   # Discount the use by request key, removing the SF_USED record
   # entirely if it's now moot.
   my $request_key = $request->[RQ_CONN_KEY];
@@ -430,6 +524,9 @@ sub _ka_conn_success {
   my $wheel_rec = delete $self->[SF_WHEELS]{$wheel_id};
   my $request   = $wheel_rec->[WHEEL_REQUEST];
 
+  # remove the wheel-to-request index
+  delete $self->[SF_REQ_INDEX]{$request->[RQ_ID]};
+
   # Remove the SF_USED placeholder, add in the socket, and store it
   # properly.
   my $used = delete $self->[SF_USED]{$wheel_id};
@@ -641,6 +738,7 @@ sub _ka_shutdown {
   foreach my $host (keys %{$heap->{resolve}}) {
     DEBUG and warn "SHT: Shutting down resolver requests for $host";
     foreach my $request (@{$heap->{resolve}{$host}}) {
+      next if $request eq "cancelled";
       $self->_shutdown_request($kernel, $request);
     }
   }
@@ -669,6 +767,9 @@ sub _shutdown_request {
   if (defined $request->[RQ_WHEEL_ID]) {
     DEBUG and warn "SHT: Shutting down resolver wheel $request->[RQ_TIMER_ID]";
     delete $self->[SF_WHEELS]{$request->[RQ_WHEEL_ID]};
+    
+    # remove the wheel-to-request index
+    delete $self->[SF_REQ_INDEX]{$request->[RQ_ID]};
   }
 
   if (defined $request->[RQ_SESSION]) {
@@ -748,7 +849,18 @@ sub _ka_dns_response {
   # being called multiple times, so we should be safe there, too.
   # Leave the die() in until coverage == 100%, so we can catch
   # possible bugs in the future.
-  die "!!!: Unexpectedly undefined requests" unless defined $requests;
+  #
+  # If it says "cancelled", it was.
+  return unless defined $requests;
+  if (defined $requests) {
+    if (ref $requests eq 'ARRAY') {
+        return if grep { $_ eq 'cancelled' } @$requests;
+    } else {
+      die "!!!: got a defined requests but NOT an arrayref";
+    }
+  } else {
+     die "!!!: Unexpectedly undefined requests";
+  }
 
   # No response.  This is an error.  Cancel all requests for the
   # address.  Tell everybody that their requests timed out.
@@ -809,7 +921,7 @@ sub _ka_add_to_queue {
   return if (
     ($self->[SF_USED_EACH]{$conn_key} || 0) >= $self->[SF_MAX_HOST]
   );
-
+  
   # Wake the session up, and return nothing, signifying sound and fury
   # yet to come.
   DEBUG and warn "posting wakeup for $conn_key";
